#!/bin/bash
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# shutdown barbican service

source $GRENADE_DIR/grenaderc
source $GRENADE_DIR/functions

# We need base DevStack functions for this
source $BASE_DEVSTACK_DIR/functions
source $BASE_DEVSTACK_DIR/stackrc # needed for status directory
source $BASE_DEVSTACK_DIR/lib/tls
source $BASE_DEVSTACK_DIR/lib/apache
source $BASE_DEVSTACK_DIR/lib/nova

BARBICAN_DEVSTACK_DIR=$(dirname "$0")/..
source $BARBICAN_DEVSTACK_DIR/lib/barbican
set -o xtrace

stop_barbican

# After shutdown barbican service, it needs to add an1 option to nova.conf
# The option will useful for running barbican tempest test
echo "Update nova.conf file with this option"
echo "enable image signature verification in Nova in target version only"
echo "[glance]"
echo "verify_glance_signatures = True"
iniset $NOVA_CONF glance verify_glance_signatures "True"
