# Copyright 2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
A filter middleware that just outputs to logs, for instructive/sample
purposes only.
"""

from microversion_parse import middleware as microversion_middleware

from barbican.api.controllers import versions
from barbican.api import middleware
from barbican.common import utils

LOG = utils.getLogger(__name__)


class MicroversionMiddleware(
        microversion_middleware.MicroversionMiddleware,
        middleware.Middleware):
    def __init__(self, app):
        super(MicroversionMiddleware, self).__init__(
            app, 'key-manager', versions.VERSIONS)
